@@ -27,10 +27,12 @@ use strict;
 use Cwd;
 use File::Basename;
 use File::Copy;
+use File::Temp qw/tempdir/;
 use filetest 'access';
 use Getopt::Long;
 use lib '/usr/share/devscripts';
 use Devscripts::Versort;
+use Text::ParseWords;
 BEGIN {
     eval { require LWP::UserAgent; };
     if ($@) {
@@ -65,6 +67,8 @@ sub dehs_msg ($);
 sub dehs_warn ($);
 sub dehs_die ($);
 sub dehs_output ();
+sub quoted_regex_replace ($);
+sub safe_replace ($$);
 
 sub usage {
     print <<"EOF";
@@ -89,7 +93,8 @@ Options:
     --symlink      Make an orig.tar.gz symlink to downloaded file (default)
     --rename       Rename to orig.tar.gz instead of symlinking
                    (Both will use orig.tar.bz2 if appropriate)
-    --repack       Repack downloaded archives from orig.tar.bz2 to orig.tar.gz
+    --repack       Repack downloaded archives from orig.tar.bz2 or orig.zip to
+                   orig.tar.gz
                    (does nothing if downloaded archive orig.tar.gz)
     --no-symlink   Don\'t make symlink or rename
     --verbose      Give verbose output
@@ -103,13 +108,19 @@ Options:
                    What constitutes a matching directory name; REGEX is
                    a Perl regular expression; the string \`PACKAGE\' will
                    be replaced by the package name; see manpage for details
-                   (default: 'PACKAGE(-.*)?')
+                   (default: 'PACKAGE(-.+)?')
     --watchfile FILE
                    Specify the watchfile rather than using debian/watch;
                    no directory traversing will be done in this case
     --upstream-version VERSION
                    Specify the current upstream version in use rather than
                    parsing debian/changelog to determine this
+    --download-version VERSION
+                   Specify the version which the upstream release must
+                   match in order to be considered, rather than using the
+                   release with the highest version
+    --download-current-version
+                   Download the currently packaged version
     --package PACKAGE
                    Specify the package name rather than examining
                    debian/changelog; must use --upstream-version and
@@ -150,13 +161,14 @@ our $passive = 'default';
 
 my $destdir = "..";
 my $download = 1;
+my $download_version;
 my $force_download = 0;
 my $report = 0; # report even on up-to-date packages?
-my $repack = 0; # repack .tar.bz2 to .tar.gz
+my $repack = 0; # repack .tar.bz2 or .zip to .tar.gz
 my $symlink = 'symlink';
 my $verbose = 0;
 my $check_dirname_level = 1;
-my $check_dirname_regex = 'PACKAGE(-.*)?';
+my $check_dirname_regex = 'PACKAGE(-.+)?';
 my $dehs = 0;
 my %dehs_tags;
 my $dehs_end_output = 0;
@@ -179,8 +191,9 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
 		       'USCAN_VERBOSE' => 'no',
 		       'USCAN_DEHS_OUTPUT' => 'no',
 		       'USCAN_USER_AGENT' => '',
+		       'USCAN_REPACK' => 'no',
 		       'DEVSCRIPTS_CHECK_DIRNAME_LEVEL' => 1,
-		       'DEVSCRIPTS_CHECK_DIRNAME_REGEX' => 'PACKAGE(-.*)?',
+		       'DEVSCRIPTS_CHECK_DIRNAME_REGEX' => 'PACKAGE(-.+)?',
 		       );
     my %config_default = %config_vars;
 
@@ -214,6 +227,8 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
 	or $config_vars{'USCAN_VERBOSE'}='no';
     $config_vars{'USCAN_DEHS_OUTPUT'} =~ /^(yes|no)$/
 	or $config_vars{'USCAN_DEHS_OUTPUT'}='no';
+    $config_vars{'USCAN_REPACK'} =~ /^(yes|no)$/
+	or $config_vars{'USCAN_REPACK'}='no';
     $config_vars{'DEVSCRIPTS_CHECK_DIRNAME_LEVEL'} =~ /^[012]$/
 	or $config_vars{'DEVSCRIPTS_CHECK_DIRNAME_LEVEL'}=1;
 
@@ -237,7 +252,8 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
     $check_dirname_level = $config_vars{'DEVSCRIPTS_CHECK_DIRNAME_LEVEL'};
     $check_dirname_regex = $config_vars{'DEVSCRIPTS_CHECK_DIRNAME_REGEX'};
     $user_agent_string = $config_vars{'USCAN_USER_AGENT'}
-	if defined $config_vars{'USCAN_USER_AGENT'};
+	if $config_vars{'USCAN_USER_AGENT'};
+    $repack = $config_vars{'USCAN_REPACK'} eq 'yes' ? 1 : 0;
 }
 
 # Now read the command line arguments
@@ -246,12 +262,15 @@ my ($opt_h, $opt_v, $opt_destdir, $opt_download, $opt_force_download,
     $opt_report, $opt_passive, $opt_symlink, $opt_repack);
 my ($opt_verbose, $opt_ignore, $opt_level, $opt_regex, $opt_noconf);
 my ($opt_package, $opt_uversion, $opt_watchfile, $opt_dehs, $opt_timeout);
+my $opt_download_version;
 my $opt_user_agent;
+my $opt_download_current_version;
 
 GetOptions("help" => \$opt_h,
 	   "version" => \$opt_v,
 	   "destdir=s" => \$opt_destdir,
 	   "download!" => \$opt_download,
+	   "download-version=s" => \$opt_download_version,
 	   "force-download" => \$opt_force_download,
 	   "report" => sub { $opt_download = 0; },
 	   "report-status" => sub { $opt_download = 0; $opt_report = 1; },
@@ -273,6 +292,7 @@ GetOptions("help" => \$opt_h,
 	   "useragent=s" => \$opt_user_agent,
 	   "noconf" => \$opt_noconf,
 	   "no-conf" => \$opt_noconf,
+	   "download-current-version" => \$opt_download_current_version,
 	   )
     or die "Usage: $progname [options] [directories]\nRun $progname --help for more details\n";
 
@@ -296,6 +316,7 @@ $symlink = $opt_symlink if defined $opt_symlink;
 $verbose = $opt_verbose if defined $opt_verbose;
 $dehs = $opt_dehs if defined $opt_dehs;
 $user_agent_string = $opt_user_agent if defined $opt_user_agent;
+$download_version = $opt_download_version if defined $opt_download_version;
 if ($dehs) {
     $SIG{'__WARN__'} = \&dehs_warn;
     $SIG{'__DIE__'} = \&dehs_die;
@@ -316,8 +337,8 @@ if (defined $opt_level) {
 $check_dirname_regex = $opt_regex if defined $opt_regex;
 
 if (defined $opt_package) {
-    die "$progname: --package requires the use of --upstream-version and --watchfile\nas well; run $progname --help for more details\n"
-	unless defined $opt_uversion and defined $opt_watchfile;
+    die "$progname: --package requires the use of --watchfile\nas well; run $progname --help for more details\n"
+	unless defined $opt_watchfile;
     $download = -$download unless defined $opt_download;
 }
 
@@ -349,7 +370,30 @@ else { $passive = undef; }
 #   else { delete $ENV{'FTP_PASSIVE'}; }
 # to restore $ENV{'FTP_PASSIVE'} to what it was at this point
 
-my $user_agent = LWP::UserAgent->new(env_proxy => 1);
+# dummy subclass used to store all the redirections for later use
+package LWP::UserAgent::UscanCatchRedirections;
+
+use base 'LWP::UserAgent';
+
+my @uscan_redirections;
+
+sub redirect_ok {
+    my $self = shift;
+    my ($request) = @_;
+    if ($self->SUPER::redirect_ok(@_)) {
+	push @uscan_redirections, $request->uri;
+	return 1;
+    }
+    return 0;
+}
+
+sub get_redirections {
+    return \@uscan_redirections;
+}
+
+package main;
+
+my $user_agent = LWP::UserAgent::UscanCatchRedirections->new(env_proxy => 1);
 $user_agent->timeout($timeout);
 $user_agent->agent($user_agent_string);
 
@@ -414,7 +458,7 @@ if (defined $opt_watchfile) {
     # Are there any warnings to give if we're using dehs?
     $dehs_end_output=1;
     dehs_output if $dehs;
-    exit 0;
+    exit ($found ? 0 : 1);
 }
 
 # Otherwise we're scanning for watchfiles
@@ -505,6 +549,8 @@ for my $dir (@dirs) {
     }
 }
 
+warn "$progname: no watch file found\n" if (@debdirs == 0 and $report);
+
 # Was there a --uversion option?
 if (defined $opt_uversion) {
     if (@debdirs == 1) {
@@ -641,13 +687,19 @@ sub process_watchline ($$$$$$)
 
     my $origline = $line;
     my ($base, $site, $dir, $filepattern, $pattern, $lastversion, $action);
+    my $basedir;
+    my (@patterns, @sites, @redirections, @basedirs);
     my %options = ();
 
     my ($request, $response);
     my ($newfile, $newversion);
     my $style='new';
     my $urlbase;
+    my $headers = HTTP::Headers->new;
 
+    # Comma-separated list of features that sites being queried might
+    # want to be aware of
+    $headers->header('X-uscan-features' => 'enhanced-matching');
     %dehs_tags = ('package' => $pkg);
 
     if ($watch_version == 1) {
@@ -732,6 +784,11 @@ sub process_watchline ($$$$$$)
 	    (undef, $lastversion, $action) = split ' ', $line, 3;
 	}
 
+	if ((!$lastversion or $lastversion eq 'debian') and not defined $pkg_version) {
+	    warn "$progname warning: Unable to determine current version\n  in $watchfile, skipping:\n  $line\n";
+	    return 1;
+	}
+
 	# Check all's OK
 	if ($filepattern !~ /\(.*\)/) {
 	    warn "$progname warning: Filename pattern missing version delimiters ()\n  in $watchfile, skipping:\n  $line\n";
@@ -744,7 +801,10 @@ sub process_watchline ($$$$$$)
 	}
 
 	# Handle sf.net addresses specially
-	$base =~ s%^http://sf\.net/%http://qa.debian.org/watch/sf.php/%;
+	if ($base =~ m%^http://sf\.net/%) {
+	    $base =~ s%^http://sf\.net/%http://qa.debian.org/watch/sf.php/%;
+	    $filepattern .= '(?:\?.*)?';
+	}
 	if ($base =~ m%^(\w+://[^/]+)%) {
 	    $site = $1;
 	} else {
@@ -759,17 +819,47 @@ sub process_watchline ($$$$$$)
 	# We're going to make the pattern
 	# (?:(?:http://site.name)?/dir/path/)?base_pattern
 	# It's fine even for ftp sites
-	my $basedir = $base;
+	$basedir = $base;
 	$basedir =~ s%^\w+://[^/]+/%/%;
 	$pattern = "(?:(?:$site)?" . quotemeta($basedir) . ")?$filepattern";
     }
 
+    if (! $lastversion or $lastversion eq 'debian') {
+	if (defined $pkg_version) {
+	    $lastversion=$pkg_version;
+	} else {
+	    warn "$progname warning: Unable to determine current version\n  in $watchfile, skipping:\n  $line\n";
+	    return 1;
+	}
+    }
+    # And mangle it if requested
+    my $mangled_lastversion;
+    $mangled_lastversion = $lastversion;
+    foreach my $pat (@{$options{'dversionmangle'}}) {
+	if (! safe_replace(\$mangled_lastversion, $pat)) {
+	    warn "$progname: In $watchfile, potentially"
+	      . " unsafe or malformed dversionmangle"
+	      . " pattern:\n  '$pat'"
+	      . " found. Skipping watchline\n"
+	      . "  $line\n";
+	    return 1;
+	}
+    }
+    if($opt_download_current_version) {
+	$download_version = $mangled_lastversion;
+	$force_download = 1;
+    }
+
     # Check all's OK
     if ($pattern !~ /\(.*\)/) {
 	warn "$progname warning: Filename pattern missing version delimiters ()\n  in $watchfile, skipping:\n  $line\n";
 	return 1;
     }
 
+    push @patterns, $pattern;
+    push @sites, $site;
+    push @basedirs, $basedir;
+
     # What is the most recent file, based on the filenames?
     # We first have to find the candidates, then we sort them using
     # Devscripts::Versort::versort
@@ -778,13 +868,40 @@ sub process_watchline ($$$$$$)
 	    die "$progname: you must have the libcrypt-ssleay-perl package installed\nto use https URLs\n";
 	}
 	print STDERR "$progname debug: requesting URL $base\n" if $debug;
-	$request = HTTP::Request->new('GET', $base);
+	$request = HTTP::Request->new('GET', $base, $headers);
 	$response = $user_agent->request($request);
 	if (! $response->is_success) {
 	    warn "$progname warning: In watchfile $watchfile, reading webpage\n  $base failed: " . $response->status_line . "\n";
 	    return 1;
 	}
 
+	@redirections = @{$user_agent->get_redirections};
+	
+	print STDERR "$progname debug: redirections: @redirections\n"
+	    if $debug;
+
+	foreach my $_redir (@redirections) {
+	    my $base_dir = $_redir;
+	    
+	    $base_dir =~ s%^\w+://[^/]+/%/%;
+	    if ($_redir =~ m%^(\w+://[^/]+)%) {
+		my $base_site = $1;
+
+		push @patterns, "(?:(?:$base_site)?" . quotemeta($base_dir) . ")?$filepattern";
+		push @sites, $base_site;
+		push @basedirs, $base_dir;
+
+		# remove the filename, if any
+		my $base_dir_orig = $base_dir;
+		$base_dir =~ s%/[^/]*$%/%;
+		if ($base_dir ne $base_dir_orig) {
+		    push @patterns, "(?:(?:$base_site)?" . quotemeta($base_dir) . ")?$filepattern";
+		    push @sites, $base_site;
+		    push @basedirs, $base_dir;
+		}
+	    }
+	}
+
 	my $content = $response->content;
 	print STDERR "$progname debug: received content:\n$content\[End of received content]\n"
 	    if $debug;
@@ -795,7 +912,7 @@ sub process_watchline ($$$$$$)
 	# Strip comments
 	$content =~ s/<!-- .*?-->//sg;
 	# Is there a base URL given?
-	if ($content =~ /<\s*base\s+[^>]*href\s*=\s*([\"\'])(.*?)\1/) {
+	if ($content =~ /<\s*base\s+[^>]*href\s*=\s*([\"\'])(.*?)\1/i) {
 	    # Ensure it ends with /
 	    $urlbase = "$2/";
 	    $urlbase =~ s%//$%/%;
@@ -804,26 +921,36 @@ sub process_watchline ($$$$$$)
 	    ($urlbase = $base) =~ s%/[^/]*$%/%;
 	}
 
-	print STDERR "$progname debug: matching pattern $pattern\n" if $debug;
+	print STDERR "$progname debug: matching pattern(s) @patterns\n" if $debug;
 	my @hrefs;
-	while ($content =~ m/<\s*a\s+[^>]*href\s*=\s*([\"\'])(.*?)\1/gi) {
+	while ($content =~ m/<\s*a\s+[^>]*href\s*=\s*([\"\'])(.*?)\1/sgi) {
 	    my $href = $2;
-	    if ($href =~ m&^$pattern$&) {
-		if ($watch_version == 2) {
-		    # watch_version 2 only recognised one group; the code
-		    # below will break version 2 watchfiles with a construction
-		    # such as file-([\d\.]+(-\d+)?) (bug #327258)
-		    push @hrefs, [$1, $href];
-		} else {
-		    # need the map { ... } here to handle cases of (...)?
-		    # which may match but then return undef values
-		    my $mangled_version =
-			join(".", map { $_ if defined($_) }
-			     $href =~ m&^$pattern$&);
-		    foreach my $pat (@{$options{'uversionmangle'}}) {
-			eval "\$mangled_version =~ $pat;";
+	    $href =~ s/\n//g;
+	    foreach my $_pattern (@patterns) {
+		if ($href =~ m&^$_pattern$&) {
+		    if ($watch_version == 2) {
+			# watch_version 2 only recognised one group; the code
+			# below will break version 2 watchfiles with a construction
+			# such as file-([\d\.]+(-\d+)?) (bug #327258)
+			push @hrefs, [$1, $href];
+		    } else {
+			# need the map { ... } here to handle cases of (...)?
+			# which may match but then return undef values
+			my $mangled_version =
+			    join(".", map { $_ if defined($_) }
+			 	$href =~ m&^$_pattern$&);
+			foreach my $pat (@{$options{'uversionmangle'}}) {
+			    if (! safe_replace(\$mangled_version, $pat)) {
+				warn "$progname: In $watchfile, potentially"
+			 	 . " unsafe or malformed uversionmangle"
+				  . " pattern:\n  '$pat'"
+				  . " found. Skipping watchline\n"
+				  . "  $line\n";
+				return 1;
+			    }
+			}
+			push @hrefs, [$mangled_version, $href];
 		    }
-		    push @hrefs, [$mangled_version, $href];
 		}
 	    }
 	}
@@ -832,8 +959,19 @@ sub process_watchline ($$$$$$)
 		print "-- Found the following matching hrefs:\n";
 		foreach my $href (@hrefs) { print "     $$href[1]\n"; }
 	    }
-	    @hrefs = Devscripts::Versort::versort(@hrefs);
-	    ($newversion, $newfile) = @{$hrefs[0]};
+	    if (defined $download_version) {
+		my @vhrefs = grep { $$_[0] eq $download_version } @hrefs;
+		if (@vhrefs) {
+		    ($newversion, $newfile) = @{$vhrefs[0]};
+		} else {
+		    warn "$progname warning: In $watchfile no matching hrefs for version $download_version"
+			. " in watch line\n  $line\n";
+		    return 1;
+		}
+	    } else {
+		@hrefs = Devscripts::Versort::versort(@hrefs);
+		($newversion, $newfile) = @{$hrefs[0]};
+	    }
 	} else {
 	    warn "$progname warning: In $watchfile,\n  no matching hrefs for watch line\n  $line\n";
 	    return 1;
@@ -876,22 +1014,36 @@ sub process_watchline ($$$$$$)
 	# that we can target our search correctly
 	if ($content =~ /<\s*a\s+[^>]*href/i) {
 	    while ($content =~ 
-		m/(?:<\s*a\s+[^>]*href\s*=\s*\")($pattern)\"/gi) {
+		m/(?:<\s*a\s+[^>]*href\s*=\s*\")((?-i)$pattern)\"/gi) {
 		my $file = $1;
 		my $mangled_version = join(".", $file =~ m/^$pattern$/);
 		foreach my $pat (@{$options{'uversionmangle'}}) {
-		    eval "\$mangled_version =~ $pat;";
+		    if (! safe_replace(\$mangled_version, $pat)) {
+			warn "$progname: In $watchfile, potentially"
+			  . " unsafe or malformed uversionmangle"
+			  . " pattern:\n  '$pat'"
+			  . " found. Skipping watchline\n"
+			  . "  $line\n";
+			return 1;
+		    }
 		}
 		push @files, [$mangled_version, $file];
 	    }
 	} else {
 	    # they all look like:
 	    # info info ... info filename [ -> linkname]
-	    while ($content =~ m/($filepattern)(\s+->\s+\S+)?$/mgi) {
+	    while ($content =~ m/\s($filepattern)(\s+->\s+\S+)?$/mg) {
 		my $file = $1;
 		my $mangled_version = join(".", $file =~ m/^$filepattern$/);
 		foreach my $pat (@{$options{'uversionmangle'}}) {
-		    eval "\$mangled_version =~ $pat;";
+		    if (! safe_replace(\$mangled_version, $pat)) {
+			warn "$progname: In $watchfile, potentially"
+			  . " unsafe or malformed uversionmangle"
+			  . " pattern:\n  '$pat'"
+			  . " found. Skipping watchline\n"
+			  . "  $line\n";
+			return 1;
+		    }
 		}
 		push @files, [$mangled_version, $file];
 	    }
@@ -902,8 +1054,19 @@ sub process_watchline ($$$$$$)
 		print "-- Found the following matching files:\n";
 		foreach my $file (@files) { print "     $$file[1]\n"; }
 	    }
-	    @files = Devscripts::Versort::versort(@files);
-	    ($newversion, $newfile) = @{$files[0]};
+	    if (defined $download_version) {
+		my @vfiles = grep { $$_[0] eq $download_version } @files;
+		if (@vfiles) {
+		    ($newversion, $newfile) = @{$vfiles[0]};
+		} else {
+		    warn "$progname warning: In $watchfile no matching files for version $download_version"
+			. " in watch line\n  $line\n";
+		    return 1;
+		}
+	    } else {
+		@files = Devscripts::Versort::versort(@files);
+		($newversion, $newfile) = @{$files[0]};
+	    }
 	} else {
 	    warn "$progname warning: In $watchfile no matching files for watch line\n  $line\n";
 	    return 1;
@@ -936,7 +1099,14 @@ EOF
         $newfile_base=$newfile;
     }
     foreach my $pat (@{$options{'filenamemangle'}}) {
-	eval "\$newfile_base =~ $pat;";
+	if (! safe_replace(\$newfile_base, $pat)) {
+	    warn "$progname: In $watchfile, potentially"
+	      . " unsafe or malformed filenamemangle"
+	      . " pattern:\n  '$pat'"
+	      . " found. Skipping watchline\n"
+	      . "  $line\n";
+		return 1;
+	}
     }
     # Remove HTTP header trash
     if ($site =~ m%^https?://%) {
@@ -946,15 +1116,7 @@ EOF
 	    $newfile_base = "$pkg-$newversion.download";
 	}
     }
-    if (! $lastversion or $lastversion eq 'debian') {
-	$lastversion=$pkg_version;
-    }
-    # And mangle it if requested
-    my $mangled_lastversion = $lastversion;
-    foreach my $pat (@{$options{'dversionmangle'}}) {
-	eval "\$mangled_lastversion =~ $pat;";
-    }
-
+    
     # So what have we got to report now?
     my $upstream_url;
     # Upstream URL?  Copying code from below - ugh.
@@ -965,18 +1127,64 @@ EOF
 	}
 	# absolute filename?
 	elsif ($newfile =~ m%^/%) {
-	    $upstream_url = "$site$newfile";
+	    # Were there any redirections? If so try using those first
+	    if ($#patterns > 0) {
+		# replace $site here with the one we were redirected to
+		foreach my $index (0 .. $#patterns) {
+		    if ("$sites[$index]$newfile" =~ m&^$patterns[$index]$&) {
+			$upstream_url = "$sites[$index]$newfile";
+			last;
+		    }
+		}
+		if (!defined($upstream_url)) {
+		    if ($debug) {
+			warn "$progname warning: Unable to determine upstream url from redirections,\n" .
+			    "defaulting to using site specified in watchfile\n";
+		    }
+		    $upstream_url = "$sites[0]$newfile";
+		}
+	    } else {
+		$upstream_url = "$sites[0]$newfile";
+	    }
 	}
 	# relative filename, we hope
 	else {
-	    $upstream_url = "$urlbase$newfile";
+	    # Were there any redirections? If so try using those first
+	    if ($#patterns > 0) {
+		# replace $site here with the one we were redirected to
+		foreach my $index (0 .. $#patterns) {
+		    # skip unless the basedir looks like a directory
+		    next unless $basedirs[$index] =~ m%/$%;
+		    my $nf = "$basedirs[$index]$newfile";
+		    if ("$sites[$index]$nf" =~ m&^$patterns[$index]$&) {
+			$upstream_url = "$sites[$index]$nf";
+			last;
+		    }
+		}
+		if (!defined($upstream_url)) {
+		    if ($debug) {
+			warn "$progname warning: Unable to determine upstream url from redirections,\n" .
+			    "defaulting to using site specified in watchfile\n";
+		    }
+		    $upstream_url = "$urlbase$newfile";
+		}
+	    } else {
+		$upstream_url = "$urlbase$newfile";
+	    }
 	}
 
 	# mangle if necessary
 	$upstream_url =~ s/&amp;/&/g;
 	if (exists $options{'downloadurlmangle'}) {
 	    foreach my $pat (@{$options{'downloadurlmangle'}}) {
-		eval "\$upstream_url =~ $pat;";
+		if (! safe_replace(\$upstream_url, $pat)) {
+		    warn "$progname: In $watchfile, potentially"
+		      . " unsafe or malformed downloadurlmangle"
+		      . " pattern:\n  '$pat'"
+		      . " found. Skipping watchline\n"
+		      . "  $line\n";
+		    return 1;
+		}
 	    }
 	}
     }
@@ -992,7 +1200,7 @@ EOF
 
     # Can't just use $lastversion eq $newversion, as then 0.01 and 0.1
     # compare different, whereas they are treated as equal by dpkg
-    if (system("dpkg --compare-versions '$mangled_lastversion' eq '$newversion'") == 0) {
+    if (system("dpkg", "--compare-versions", "'$mangled_lastversion'", "eq", "'$newversion'") == 0) {
 	if ($verbose or ($download == 0 and $report and ! $dehs)) {
 	    print $pkg_report_header;
 	    $pkg_report_header = '';
@@ -1018,18 +1226,24 @@ EOF
 
     # We use dpkg's rules to determine whether our current version
     # is newer or older than the remote version.
-    if (system("dpkg --compare-versions '$mangled_lastversion' gt '$newversion'") == 0) {
-        if ($verbose) {
-	    print " => remote site does not even have current version\n";
-	} elsif ($dehs) {
-	    $dehs_tags{'status'} = "Debian version newer than remote site";
+    if (!defined $download_version) {
+	if (system("dpkg", "--compare-versions", "'$mangled_lastversion'", "gt", "'$newversion'") == 0) {
+	    if ($verbose) {
+		print " => remote site does not even have current version\n";
+	    } elsif ($dehs) {
+		$dehs_tags{'status'} = "Debian version newer than remote site";
+	    } else {
+		print "$pkg: remote site does not even have current version\n";
+	    }
+	    return 0;
 	} else {
-	    print "$pkg: remote site does not even have current version\n";
+	    # There's a newer upstream version available, which may already
+	    # be on our system or may not be
+	    $found++;
 	}
-        return 0;
     } else {
-	# There's a newer upstream version available, which may already
-	# be on our system or may not be
+	# Flag that we found a newer upstream version, so that the exit status
+	# is set correctly
 	$found++;
     }
 
@@ -1130,12 +1344,35 @@ EOF
     if ($repack and $newfile_base =~ /^(.*)\.(tar\.bz2|tbz2?)$/) {
 	print "-- Repacking from bzip2 to gzip\n" if $verbose;
 	my $newfile_base_gz = "$1.tar.gz";
-	system("bunzip2 -c $destdir/$newfile_base | gzip > $destdir/$newfile_base_gz") == 0
+	system("bunzip2 -c $destdir/$newfile_base | gzip -n -9 > $destdir/$newfile_base_gz") == 0
 	  or die "repacking from bzip2 to gzip failed\n";
 	unlink "$destdir/$newfile_base";
 	$newfile_base = $newfile_base_gz;
     }
 
+    if ($repack and $newfile_base =~ /^(.*)\.(tar\.lzma|tlzma?)$/) {
+	print "-- Repacking from lzma to gzip\n" if $verbose;
+	my $newfile_base_gz = "$1.tar.gz";
+	system("lzma -cd $destdir/$newfile_base | gzip -n -9 > $destdir/$newfile_base_gz") == 0
+	  or die "repacking from lzma to gzip failed\n";
+	unlink "$destdir/$newfile_base";
+	$newfile_base = $newfile_base_gz;
+    }
+
+    if ($repack and $newfile_base =~ /^(.*)\.zip$/) {
+	print "-- Repacking from zip to .tar.gz\n" if $verbose;
+
+	system('command -v unzip >/dev/null 2>&1') >> 8 == 0
+	  or die("unzip binary not found. You need to install the package unzip to be able to repack .zip upstream archives.\n");
+
+	my $newfile_base_gz = "$1.tar.gz";
+	my $tempdir = tempdir ( "uscanXXXX", TMPDIR => 1, CLEANUP => 1 );
+	system("unzip -q -d $tempdir $destdir/$newfile_base; GZIP=-9 tar -C $tempdir -czf $destdir/$newfile_base_gz .") == 0 
+	  or die("Repacking from zip to tar.gz failed\n");
+	unlink "$destdir/$newfile_base";
+	$newfile_base = $newfile_base_gz;
+    }
+
     if ($newfile_base =~ /\.(tar\.gz|tgz|tar\.bz2|tbz2?)$/) {
 	my $filetype = `file $destdir/$newfile_base`;
 	$filetype =~ s%^\.\./\Q$newfile_base\E: %%;
@@ -1214,6 +1451,7 @@ EOF
     # Do whatever the user wishes to do
     if ($action) {
 	my $usefile = "$destdir/$newfile_base";
+	my @cmd = shellwords($action);
 	if ($symlink =~ /^(symlink|rename)$/
 	    and $newfile_base =~ /\.(tar\.gz|tgz)$/) {
 	    $usefile = "$destdir/${pkg}_${newversion}.orig.tar.gz";
@@ -1225,27 +1463,22 @@ EOF
 
 	# Any symlink requests are already handled by uscan
 	if ($action =~ /^uupdate(\s|$)/) {
-	    $action =~ s/^uupdate/uupdate --no-symlink/;
+	    push @cmd, "--no-symlink";
 	}
 
 	if ($watch_version > 1) {
-	    print "-- Executing user specified script\n     $action --upstream-version $newversion $newfile_base" if $verbose;
-	    if ($dehs) {
-		my $msg = "Executing user specified script: $action --upstream-version $newversion $newfile_base; output:\n";
-		$msg .= `$action --upstream-version $newversion $usefile 2>&1`;
-		dehs_msg($msg);
-	    } else {
-		system("$action --upstream-version $newversion $usefile");
-	    }
+	    push @cmd, ("--upstream-version", "$newversion", "$usefile");
 	} else {
-	    print "-- Executing user specified script $action $newfile_base $newversion" if $verbose;
-	    if ($dehs) {
-		my $msg = "Executing user specified script: $action $newfile_base $newversion; output:\n";
-		$msg .= `$action $usefile $newversion 2>&1`;
-		dehs_msg($msg);
-	    } else {
-		system("$action $usefile $newversion");
-	    }
+	    push @cmd, ("$usefile", "$newversion");
+	}
+	my $actioncmd = join(" ", @cmd);
+	print "-- Executing user specified script\n     $actioncmd\n" if $verbose;
+	if ($dehs) {
+	    my $msg = "Executing user specified script: $actioncmd; output:\n";
+	    $msg .= `$actioncmd 2>&1`;
+	    dehs_msg($msg);
+	} else {
+	    system(@cmd);
 	}
     }
 
@@ -1373,9 +1606,9 @@ sub newest_dir ($$$$$) {
 
 	# We separate out HTMLised listings from standard listings, so
 	# that we can target our search correctly
-	if (/<\s*a\s+[^>]*href/i) {
+	if ($content =~ /<\s*a\s+[^>]*href/i) {
 	    while ($content =~ 
-		m/(?:<\s*a\s+[^>]*href\s*=\s*\")($pattern)\"/gi) {
+		m/(?:<\s*a\s+[^>]*href\s*=\s*\")((?-i)$pattern)\"/gi) {
 		my $dir = $1;
 		my $mangled_version = join(".", $dir =~ m/^$pattern$/);
 		push @dirs, [$mangled_version, $dir];
@@ -1383,7 +1616,7 @@ sub newest_dir ($$$$$) {
 	} else {
 	    # they all look like:
 	    # info info ... info filename [ -> linkname]
-	    while ($content =~ m/($pattern)(\s+->\s+\S+)?$/mgi) {
+	    while ($content =~ m/($pattern)(\s+->\s+\S+)?$/mg) {
 		my $dir = $1;
 		my $mangled_version = join(".", $dir =~ m/^$pattern$/);
 		push @dirs, [$mangled_version, $dir];
@@ -1398,7 +1631,7 @@ sub newest_dir ($$$$$) {
 	    my ($newversion, $newdir) = @{$dirs[0]};
 	    return $newdir;
 	} else {
-	    warn "$progname warning: In $watchfile no matching dirs for pattern\n  $site$base$pattern\n";
+	    warn "$progname warning: In $watchfile no matching dirs for pattern\n  $base$pattern\n";
 	    return '';
 	}
     }
@@ -1534,3 +1767,211 @@ sub dehs_output ()
     # Don't repeat output
     %dehs_tags = ();
 }
+
+sub quoted_regex_parse($) {
+    my $pattern = shift;
+    my %closers = ('{', '}', '[', ']', '(', ')', '<', '>');
+
+    $pattern =~ /^(s|tr|y)(.)(.*)$/;
+    my ($sep, $rest) = ($2, $3 || '');
+    my $closer = $closers{$sep};
+
+    my $parsed_ok = 1;
+    my $regexp = '';
+    my $replacement = '';
+    my $flags = '';
+    my $open = 1;
+    my $last_was_escape = 0;
+    my $in_replacement = 0;
+
+    for my $char (split //, $rest) {
+	if ($char eq $sep and ! $last_was_escape) {
+	    $open++;
+	    if ($open == 1) {
+		if ($in_replacement) {
+		    # Separator after end of replacement
+		    $parsed_ok = 0;
+		    last;
+		} else {
+		    $in_replacement = 1;
+		}
+	    } else {
+		if ($open > 1) {
+		    if ($in_replacement) {
+			$replacement .= $char;
+		    } else {
+			$regexp .= $char;
+		    }
+		}
+	    }
+	} elsif ($char eq $closer and ! $last_was_escape) {
+	    $open--;
+	    if ($open) {
+		if ($in_replacement) {
+		    $replacement .= $char;
+		} else {
+		    $regexp .= $char;
+		}
+	    } elsif ($open < 0) {
+		$parsed_ok = 0;
+		last;
+	    }
+	} else {
+	    if ($in_replacement) {
+		if ($open) {
+		    $replacement .= $char;
+		} else {
+		    $flags .= $char;
+		}
+	    } else {
+		$regexp .= $char;
+	    }
+	}
+	# Don't treat \\ as an escape
+	$last_was_escape = ($char eq '\\' and ! $last_was_escape);
+    }
+
+    $parsed_ok = 0 unless $in_replacement and $open == 0;
+
+    return ($parsed_ok, $regexp, $replacement, $flags);
+}
+
+sub safe_replace($$) {
+    my ($in, $pat) = @_;
+    $pat =~ s/^\s*(.*?)\s*$/$1/;
+
+    $pat =~ /^(s|tr|y)(.)/;
+    my ($op, $sep) = ($1, $2 || '');
+    my $esc = "\Q$sep\E";
+    my ($parsed_ok, $regexp, $replacement, $flags);
+
+    if ($sep eq '{' or $sep eq '(' or $sep eq '[' or $sep eq '<') {
+	($parsed_ok, $regexp, $replacement, $flags) = quoted_regex_parse($pat);
+
+	return 0 unless $parsed_ok;
+    } elsif ($pat !~ /^(?:s|tr|y)$esc((?:\\.|[^\\$esc])*)$esc((?:\\.|[^\\$esc])*)$esc([a-z]*)$/) {
+	return 0;
+    } else {
+	($regexp, $replacement, $flags) = ($1, $2, $3);
+    }
+
+    my $safeflags = $flags;
+    if ($op eq 'tr' or $op eq 'y') {
+	$safeflags =~ tr/cds//cd;
+	return 0 if $safeflags ne $flags;
+	
+	$regexp =~ s/\\(.)/$1/g;
+	$replacement =~ s/\\(.)/$1/g;
+
+	$regexp =~ s/([^-])/'\\x'  . unpack 'H*', $1/ge;
+	$replacement =~ s/([^-])/'\\x'  . unpack 'H*', $1/ge;
+
+	eval "\$\$in =~ tr<$regexp><$replacement>$flags;";
+
+	if ($@) {
+	    return 0;
+	} else {
+	    return 1;
+	}
+    } else {
+	$safeflags =~ tr/gix//cd;
+	return 0 if $safeflags ne $flags;
+
+	my $global = ($flags =~ s/g//);
+	$flags = "(?$flags)" if length $flags;
+
+	my $slashg;
+	if ($regexp =~ /(?<!\\)(\\\\)*\\G/) {
+	    $slashg = 1;
+	    # if it's not initial, it is too dangerous
+	    return 0 if $regexp =~ /^.*[^\\](\\\\)*\\G/;
+	}
+
+	# Behave like Perl and treat e.g. "\." in replacement as "."
+	# We allow the case escape characters to remain and
+	# process them later
+	$replacement =~ s/(^|[^\\])\\([^luLUE])/$1$2/g;
+
+	# Unescape escaped separator characters
+	$replacement =~ s/\\\Q$sep\E/$sep/g;
+	# If bracketing quotes were used, also unescape the
+	# closing version
+	$replacement =~ s/\\\Q}\E/}/g if $sep eq '{';
+	$replacement =~ s/\\\Q]\E/]/g if $sep eq '[';
+	$replacement =~ s/\\\Q)\E/)/g if $sep eq '(';
+	$replacement =~ s/\\\Q>\E/>/g if $sep eq '<';
+
+	# The replacement below will modify $replacement so keep
+	# a copy. We'll need to restore it to the current value if
+	# the global flag was set on the input pattern.
+	my $orig_replacement = $replacement;
+
+	my ($first, $last, $pos, $zerowidth, $matched, @captures) = (0, -1, 0);
+	while (1) {
+	    eval {
+		# handle errors due to unsafe constructs in $regexp
+		no re 'eval';
+
+		# restore position
+		pos($$in) = $pos if $pos;
+
+		if ($zerowidth) {
+		    # previous match was a zero-width match, simulate it to set
+		    # the internal flag that avoids the infinite loop
+		    $$in =~ /()/g;
+		}
+		# Need to use /g to make it use and save pos()
+		$matched = ($$in =~ /$flags$regexp/g);
+
+		if ($matched) {
+		    # save position and size of the match
+		    my $oldpos = $pos;
+		    $pos = pos($$in);
+		    ($first, $last) = ($-[0], $+[0]);
+
+		    if ($slashg) {
+			# \G in the match, weird things can happen
+			$zerowidth = ($pos == $oldpos);
+			# For example, matching without a match
+			$matched = 0 if (not defined $first
+			    or not defined $last);
+		    } else {
+			$zerowidth = ($last - $first == 0);
+		    }
+		    for my $i (0..$#-) {
+			$captures[$i] = substr $$in, $-[$i], $+[$i] - $-[$i];
+		    }
+		}
+	    };
+	    return 0 if $@;
+
+	    # No match; leave the original string  untouched but return
+	    # success as there was nothing wrong with the pattern
+	    return 1 unless $matched;
+
+	    # Replace $X
+	    $replacement =~ s/[\$\\](\d)/defined $captures[$1] ? $captures[$1] : ''/ge;
+	    $replacement =~ s/\$\{(\d)\}/defined $captures[$1] ? $captures[$1] : ''/ge;
+	    $replacement =~ s/\$&/$captures[0]/g;
+
+	    # Make \l etc escapes work
+	    $replacement =~ s/\\l(.)/lc $1/e;
+	    $replacement =~ s/\\L(.*?)(\\E|\z)/lc $1/e;
+	    $replacement =~ s/\\u(.)/uc $1/e;
+	    $replacement =~ s/\\U(.*?)(\\E|\z)/uc $1/e;
+
+	    # Actually do the replacement
+	    substr $$in, $first, $last - $first, $replacement;
+	    # Update position
+	    $pos += length($replacement) - ($last - $first);
+
+	    if ($global) {
+		$replacement = $orig_replacement;
+	    } else {
+		last;
+	    }
+ 	}
+
+	return 1;
+    }
+}
